﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  using Interpreter;

  /// <summary>
  ///     Pushes the sum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.+")]
  public class IntegerAddExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2)
        return false;

      var first = interpreter.IntegerStack.Pop();
      var second = interpreter.IntegerStack.Top;
      var result = second + first;

      interpreter.IntegerStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the difference of the top two items; that is, the second item minus the top item.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.-")]
  public class IntegerSubtractExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2)
        return false;

      var first = interpreter.IntegerStack.Pop();
      var second = interpreter.IntegerStack.Top;
      var result = second - first;

      interpreter.IntegerStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the product of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.*")]
  public class IntegerMultiplyExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2)
        return false;

      var first = interpreter.IntegerStack.Pop();
      var second = interpreter.IntegerStack.Top;
      var result = second * first;

      interpreter.IntegerStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the quotient of the top two items; that is, the second item divided by the top item.
  ///     If the top item is zero this acts as a NOOP.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER./")]
  public class IntegerDivideExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2 ||
          interpreter.IntegerStack.Top == 0)
        return false;

      var first = interpreter.IntegerStack.Pop();
      var second = interpreter.IntegerStack.Top;
      var result = second / first;

      interpreter.IntegerStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the second stack item modulo the top stack item. If the top item is zero this acts as a NOOP. The modulus is
  ///     computed as the remainder of the quotient, where the quotient has first been truncated toward negative infinity. (This is taken
  ///     from the definition for the generic MOD function in Common Lisp, which is described for example at
  ///     http://www.lispworks.com/reference/HyperSpec/Body/f_mod_r.htm.)
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.%")]
  public class IntegerModuloExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2 ||
          interpreter.IntegerStack.Top == 0)
        return false;

      var first = interpreter.IntegerStack.Pop();
      var second = interpreter.IntegerStack.Top;
      var result = second % first;

      interpreter.IntegerStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the minimum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.MIN")]
  public class IntegerMinExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2)
        return false;

      var first = interpreter.IntegerStack.Pop();
      var second = interpreter.IntegerStack.Top;
      var result = Math.Min(second, first);

      interpreter.IntegerStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the maximum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.MAX")]
  public class IntegerMaxExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2)
        return false;

      var first = interpreter.IntegerStack.Pop();
      var second = interpreter.IntegerStack.Top;
      var result = Math.Max(second, first);

      interpreter.IntegerStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.<", StackTypes.Boolean)]
  public class IntegerSmallerThanExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2)
        return false;

      var first = interpreter.IntegerStack.Top;
      var second = interpreter.IntegerStack[1];
      interpreter.IntegerStack.Remove(2);

      var result = second < first;
      interpreter.BooleanStack.Push(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.>", StackTypes.Boolean)]
  public class IntegerGreaterThanExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count < 2)
        return false;

      var first = interpreter.IntegerStack.Top;
      var second = interpreter.IntegerStack[1];
      interpreter.IntegerStack.Remove(2);

      var result = second > first;
      interpreter.BooleanStack.Push(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes 1 if the top BOOLEAN is TRUE, or 0 if the top BOOLEAN is FALSE.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.FROMBOOLEAN", StackTypes.Boolean)]
  public class IntegerFromBooleanExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.BooleanStack.Count == 0)
        return false;

      var condition = interpreter.BooleanStack.Pop();
      var value = condition ? 1 : 0;
      interpreter.IntegerStack.Push(value);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the result of truncating the top FLOAT.
  /// </summary>
  [PushExpression(StackTypes.Integer, "INTEGER.FROMFLOAT", StackTypes.Float)]
  public class IntegerFromFloatExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count == 0)
        return false;

      var value = (int)interpreter.FloatStack.Pop();
      interpreter.IntegerStack.Push(value);
      return true;
    }
  }
}