﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using System.Collections.Generic;
  using Attributes;
  using Interpreter;
  using Stack;

  /// <summary>
  ///     Pushes a copy of an indexed item "deep" in the stack onto the top of the stack, without removing the deep item.
  ///     The index is taken from the INTEGER stack, and the indexing is done after the index is removed.
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  public abstract class YankDuplicateExpression<T> : StatelessExpression {
    public bool Eval(IPushStack<T> stack, IPushStack<long> integerStack) {
      if ((stack == integerStack && integerStack.Count < 2) ||
          (stack != integerStack && (integerStack.IsEmpty || stack.IsEmpty)))
        return false;

      var index = (int)Math.Abs(integerStack.Pop() % stack.Count);
      stack.Push(stack[index]);
      return true;
    }
  }

  [PushExpression(StackTypes.Integer, "INTEGER.YANKDUP")]
  public class IntegerYankDuplicateExpression : YankDuplicateExpression<long> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return this.Eval(interpreter.IntegerStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Float, "FLOAT.YANKDUP", StackTypes.Integer)]
  public class FloatYankDuplicateExpression : YankDuplicateExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return this.Eval(interpreter.FloatStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Boolean, "BOOLEAN.YANKDUP", StackTypes.Integer)]
  public class BooleanYankDuplicateExpression : YankDuplicateExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return this.Eval(interpreter.BooleanStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Name, "NAME.YANKDUP", StackTypes.Integer)]
  public class NameYankDuplicateExpression : YankDuplicateExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return this.Eval(interpreter.NameStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.YANKDUP", StackTypes.Integer)]
  public class ExecYankDuplicateExpression : YankDuplicateExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return this.Eval(interpreter.ExecStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Code, "CODE.YANKDUP", StackTypes.Integer)]
  public class CodeYankDuplicateExpression : YankDuplicateExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return this.Eval(interpreter.CodeStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.YANKDUP", StackTypes.Integer)]
  public class CharYankDuplicateExpression : YankDuplicateExpression<char> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.CharStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.String, "STRING.YANKDUP", StackTypes.Integer)]
  public class StringYankDuplicateExpression : YankDuplicateExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.StringStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.IntegerVector, "INTEGER[].YANKDUP", StackTypes.Integer)]
  public class IntegerVectorYankDuplicateExpression : YankDuplicateExpression<List<long>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.FloatVector, "FLOAT[].YANKDUP", StackTypes.Integer)]
  public class FloatVectorYankDuplicateExpression : YankDuplicateExpression<List<double>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatVectorStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].YANKDUP", StackTypes.Integer)]
  public class BooleanVectorYankDuplicateExpression : YankDuplicateExpression<List<bool>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanVectorStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.StringVector, "STRING[].YANKDUP", StackTypes.Integer)]
  public class StringVectorYankDuplicateExpression : YankDuplicateExpression<List<string>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.StringVectorStack, interpreter.IntegerStack);
    }
  }
}